"""
为各个模块提供抽象类别
"""
import argparse
from curses import noecho
import os
from tkinter import N
# from damei.nn.api.utils import Config
from ..utils.config_loader import PyConfigLoader as Config


class AbstractModule(object):
    name = 'default_name'
    status = 'stopped'
    description = 'default_description'
    # default_cfg = None
    cfg_meta = None
    tag = 'cloud'  # default tag, edge or cloud, edge for edge node, cloud for center node

    def __init__(self,cfg=None, **kwargs):
        """
        初始化模块，合并配置文件
        """
        AbstractModule.status = 'ready'
        self._name = AbstractModule.name
        self.status = AbstractModule.status
        self.description = AbstractModule.description
        # self.default_cfg = AbstractModule.default_cfg
        self.default_cfg = None
        # print(default_cfg)
        # self._cfg = self._init_cfg(self.default_cfg, cfg, **kwargs)
        self._cfg = None
        self._model = None
        self.mi = None
        self.mo = None
        # print(f'xxx {self.name}:? {self.status}')
    
    def convert_config_type(self, cfg):
        """
        conver the config type from Config object, .py file argparse.Namespace to Config object
        """
        if isinstance(cfg, Config):
            return cfg
        elif isinstance(cfg, str):
            return Config.fromfile(cfg)
        elif isinstance(cfg, dict):
            return Config.from_dict(cfg)
        elif isinstance(cfg, argparse.Namespace):
            return Config.from_argparse(cfg)
        else:
            raise TypeError(f'{self.name}: {cfg} is not a valid config: {type(cfg)}')
        

    def _init_cfg(self, default_cfg, cfg, **kwargs):
        """
        init config
        :param default_cfg: .py file or Config object or ArgumentParser object 
        :param cfg: .py file or Config object or 
        """
        if default_cfg is not None:
            default_cfg = self.convert_config_type(default_cfg)
                
        if cfg is not None:
            cfg = self.convert_config_type(cfg)
            # cfg = Config.fromfile(cfg) if isinstance(cfg, str) else cfg

        if default_cfg and cfg:
            cfg = default_cfg.merge(cfg2=cfg)
        elif default_cfg:
            cfg = default_cfg
        elif cfg:
            cfg = cfg
        else:
            cfg = None
        return cfg
    
    @property
    def name(self):
        if self._name == 'default_name':
            return self.__class__.__name__
        return self._name

    @property
    def cfg(self):
        # print(self.default_cfg)
        if self._cfg is None:
            self._cfg = self._init_cfg(default_cfg=self.default_cfg, cfg=None)
        return self._cfg

    @property
    def config(self):
        """alias of cfg"""
        return self.cfg

    # @property
    # def model(self):
    #     if self._model is None:
    #         self.build_model()
    #     return self._model

    def build_model(self, *args, **kwargs):
        raise NotImplementedError(f'{self.name}.build_model() not implemented, plese check the api: "{self.__module__}"')
   
    def get_config(self):
        """alias of cfg"""
        return self.cfg

    def get_cfg(self):
        return self.cfg

    def forward(self, x, *args, **kwargs):
        return self.model(x)

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def train(self, *args, **kwargs):
        raise NotImplementedError(f'{self.name}.train() not implemented, plese check the api: "{self.__module__}"')

    def infer(self, *args, **kwargs):
        raise NotImplementedError(f'{self.name}.infer() not implemented, plese check the api: "{self.__module__}"')

    def predict(self, *args, **kwargs):
        raise NotImplementedError(f'{self.name}.predict() not implemented, plese check the api: "{self.__module__}"')

    def evaluate(self, *args, **kwargs):
        raise NotImplementedError(f'{self.name}.evaluate() not implemented, plese check the api: "{self.__module__}"')

    def before_train_run(self, *args, **kwargs):
        pass

    def before_train_epoch(self, *args, **kwargs):
        pass

    def before_train_iter(self, *args, **kwargs):
        pass

    def after_train_iter(self, *args, **kwargs):
        pass

    def after_epoch(self, *args, **kwargs):
        pass

    def after_run(self, *args, **kwargs):
        pass



"""
基础的流类
"""
import sys

import numpy as np
import collections
import copy
import damei as dm
import json
# from ..utils.config_loader import PyConfigLoader
from damei.nn.api.utils import Config

logger = dm.getLogger('stream')


class Stream(object):
    name = 'stream_name'
    status = 'stopped'
    description = 'stream description'

    def __init__(self, parent, id=None, stream_cfg=None, **kwargs):
        self.is_mono = kwargs.pop('is_mono', False)
        self.name = stream_cfg['type']
        self.status = Stream.status
        self.description = stream_cfg.get('description', Stream.description)
        self.id = id
        self._cfg = stream_cfg
        self.modules_list_config = stream_cfg['models']  # list，每个元素是dict，是模块的配置，type和cfg
        # print('初始化', parent)
        self.parent = parent  # uaii
        self._modules, self._module_cfgs, self._module_cfgs_meta = self.init_about_models()  # 未初始化的模块和对应的配置
        self._inited_modules = collections.OrderedDict()  # 已经初始化的模块
        # self.register_attrs()
        # print('xx', self.seyolov5)

    def __repr__(self):
        format_str = f"<class '{self.__class__.__name__}'> " \
                     f"(name='{self.name}', status='{self.status}', " \
                     f"include={self.include(format=list)}, " \
                     f"description='{self.description}')"
        return format_str

    def init_about_models(self):
        """
        init this stream，read configs, DO NOT init modules.
        :return:
            ordered dict of modules: key is module name, value is module class (not inited)
            ordered dict of module configs: key is module name, value is module config: Config object.
        """
        _modules = collections.OrderedDict()
        _module_cfgs = collections.OrderedDict()
        _module_cfgs_meta = collections.OrderedDict()
        for i, module in enumerate(self.modules_list_config):
            mname = module['type']  # 必须存在
            model_cfg = module.pop('cfg', None)  # 可以为无，默认为None
            module = copy.deepcopy(
                self.parent.get_module(mname, cls='MODULE', module=True))  # copy modeul from registry
            _modules[mname] = module  # 存入模块
            setattr(self, mname, module)  # 存入属性, self.module_name = module
            # cfg = PyConfigLoader(module.default_cfg)  # 这是默认配置
            # assert module.default_cfg is not None, f'"{mname}" has no default_cfg'
            cfg = Config(module.default_cfg)  # 使用模块的默认配置，的

            if model_cfg:  # 如果外部设置流的同时指定了配置文件，合并配置
                full_cfg_path = f'{self.parent.root_path}/{model_cfg}'
                # cfg2 = PyConfigLoader(cfg_file=full_cfg_path, root_path=self.parent.root_path)
                cfg2 = Config(cfg_file=full_cfg_path, root_path=self.parent.root_path)
                cfg = cfg.merge(cfg2)
            _module_cfgs[mname] = cfg

            cfg_meta_file = module.cfg_meta
            cfg_meta = Config(cfg_file=cfg_meta_file)
            _module_cfgs_meta[mname] = cfg_meta

        return _modules, _module_cfgs, _module_cfgs_meta

    def __call__(self, x, *args, **kwargs):
        """模块前馈"""
        # 需要把input封装成mi对象
        self.parent.run_stream(
            self,
            run_in_main_thread=True,
            mtask='infer',
            mi=input,
        )

    def init(self, stream_cfg=None, **kwargs):
        """初始化流内模块
        初始化流的意思的：对流内每个模块进行初始化，并配置好模块的I/O
        模块可能的状态有：stopped, ready, running.
        stopped时，直接初始化
        ready或running时，如果force_init=True, 先停止，再初始化
                如果force_init=False, 则不初始化
        """
        # TODO: 通用的monoflow和multiflow的传入stream_cfg的实现方法
        logger.info(f'Initializing ...')
        force_init = stream_cfg is not None  # 当传入stream_cfg时，强制初始化

        if self.status != 'stopped':
            if force_init:
                self.stop()
            else:
                logger.warning(f'Stream {self.name} is not stopped, skip initialization')
                return

        if self.status == 'stopped':
            models = self.models  # 所有的模型
            model_cfgs = self.model_cfgs  # 所有的配置
            for i, (mname, model) in enumerate(models.items()):
                model_cfg = model_cfgs[mname]
                is_first = i == 0
                is_last = i == len(models) - 1
                # print(model.status, force_init, self.status)
                model = model(cfg=model_cfg)

                mi = self.parent.build_io_instance('input', model_cfg) if is_first else None
                mo = self.parent.build_io_instance('output', model_cfg) if is_last else None

                if mi is not None:
                    if hasattr(model, 'mi'):
                        delattr(model, 'mi')
                    setattr(model, 'mi', mi)
                if mo is not None:
                    if hasattr(model, 'mo'):
                        delattr(model, 'mo')
                    setattr(model, 'mo', mo)

                if hasattr(self, mname):
                    delattr(self, mname)
                setattr(self, mname, model)  # 变成初始化后的模型

                self._inited_modules[mname] = model  # 把初始化后的模型存进去
        else:
            logger.warning(f'Stream {self.name} is not stopped, skip initialization')
        self.status = 'ready'

    def register_attrs(self):
        """把包含的子模块注册为本流的属性"""
        for i, module in enumerate(self.stream_models):
            mname = module['type']
            mtask = module.get('task')
            module = self.parent.get_module(mname, cls='MODULE', module=True)
            setattr(self, mname, module)

    @property
    def cfg(self):
        """流的配置文件，不包含模块的配置"""
        return self._cfg

    @property
    def models(self):
        if self.status == 'stopped':
            return self._modules
        else:
            return self._inited_modules

    @property
    def model_cfgs(self):
        return self._module_cfgs

    @property
    def model_cfgs_meta(self):
        return self._module_cfgs_meta

    @property
    def model_names(self):
        module_names = [x['type'] for x in self.modules_list_config]
        # print(self.models)
        # sys.exit(1)
        return module_names

    def get_config(self, addr='/', **kwargs):
        return self.get_cfg(addr=addr, **kwargs)

    def get_cfg_meta(self, addr='/', **kwargs):
        if self.is_mono:
            addr = f'{addr}/{self.model_names[0]}' if addr == '/' else addr

        if addr == '/':
            stream_cfg = self.cfg
            for mname in self.model_names:
                model_cfg_meta = self.model_cfgs_meta[mname]
                model_cfg_meta_dict = model_cfg_meta.to_dict()
                stream_cfg[mname] = model_cfg_meta_dict
            return Config.from_dict(stream_cfg)
        else:
            attrs = [x for x in addr.split('/') if x != '']
            assert len(attrs) >= 1, 'addr must be like /xxx/xxx/xxx'
            mname = attrs.pop(0)
            cfg_meta = self.model_cfgs_meta[mname]
            for sub_attr in attrs:
                cfg_meta = cfg_meta[sub_attr]
            return cfg_meta

    def get_cfg(self, addr='/', **kwargs):
        """
        获取配置，包含流的配置和子模块的配置，其中mono流的配置即内部模块的配置
        :param addr: 配置路径, 例如：/带代表stream自身配置，/seyolov5代表流的seyolov5模块的配置，/seyolov5/input_stream代表流的模块的input_stream的配置
        :return: 配置dict
        """

        if self.is_mono:
            addr = f'{addr}{self.model_names[0]}' if addr == '/' else addr

        if addr == '/':  # 全部配置
            stream_cfg = self.cfg
            for model_name in self.model_names:
                model_cfg = self.model_cfgs[model_name]
                model_cfg_dict = model_cfg.to_dict()
                stream_cfg[model_name] = model_cfg_dict
            # print(stream_cfg, stream_cfg.keys())
            return Config.from_dict(stream_cfg)
        else:  # 子模块配置
            attrs = [x for x in addr.split('/') if x != '']
            assert len(attrs) >= 1
            mname = attrs.pop(0)  # module name
            cfg = self.model_cfgs[mname]
            for sub_attr in attrs:
                cfg = cfg[sub_attr]
            return cfg

    def set_config(self, value=None, addr='/', **kwargs):
        return self.set_cfg(value, addr, **kwargs)

    def set_cfg(self, value=None, addr='/', **kwargs):
        """
        更新配置
        :param addr: address
        :param value: config value
        :return:
        """
        assert value is not None

        # 处理mono流的配置，当传入的addr是/时，其实是更新里面的模型的配置
        if self.is_mono:  # mono流，传入/时，更新为/model_name
            addr = f'/{self.model_names[0]}' if addr == '/' else addr

        if addr == '/':  # multi-module流, update all
            # 不能修改type, description, models配置项，对每个子模块，都要更新配置
            for model_name in self.model_names:
                model_cfg = self.model_cfgs[model_name]
                new_model_cfg = Config.from_dict(value[model_name])
                model_cfg.merge(new_model_cfg)
            # raise NotImplementedError(f"You cannot set config of the stream")
            return self.get_cfg(addr=addr, **kwargs)
        else:
            attrs = [x for x in addr.split('/') if x != '']
            assert len(attrs) >= 1
            if len(attrs) == 1:  # 即addr=/seyolov5这类情形
                if isinstance(value, dict):
                    value = Config.from_dict(value)
                else:  # 有可能本来就是Config对象，即<class 'damei.nn.uaii.utils.config_loader.PyConfigLoader'>
                    pass
            mname = attrs.pop(0)
            m = self.models[mname]
            if m.status != 'stopped':
                logger.warn(f"Please stop the stream: '{self.name}' while configure it's module.")
                return
            mcfg = self.model_cfgs[mname]
            mcfg.update_item(attrs, value)
            self.model_cfgs[mname] = mcfg
            # print(f'newcfg: {mcfg.info()}')
            return self.get_cfg(addr='/', **kwargs)

    def ps(self, *args, **kwargs):
        return self.info(*args, **kwargs)

    def get_info(self, *args, **kwargs):
        return self.info(*args, **kwargs)

    def info(self, *args, **kwargs):
        ret_fmt = kwargs.get('ret_fmt', 'string')

        info = dict()
        info['class'] = f'<class "{self.__class__.__name__}">'
        info['name'] = self.name
        info['description'] = self.description
        info['status'] = self.status
        info['is_mono'] = self.is_mono

        sub_modules = self.include()
        sub_modules_ids = [self.parent.modules.name2id_dict[x] for x in sub_modules]
        info['models'] = dict(zip(sub_modules_ids, sub_modules))

        models = self.models
        model_cfgs = self.model_cfgs
        for i, (mname, model) in enumerate(models.items()):
            # mname = module['type']
            # mtask = module.get('task', None)
            id = self.parent.modules.name2id_dict[mname]
            cfg = self.model_cfgs[mname]

            # m = self.parent.get_module(name=mname)
            # if m.status == 'stopped':
            #    cfg = PyConfigLoader(m.default_cfg)
            # else:  # ready running
            #    cfg = m.cfg  # 可能会merge之后的也是PyConfigLoader对象

            info[f'{mname} config'] = cfg.items
        # print(ret_fmt)
        # exit()
        if ret_fmt == 'string':
            format_str = self.dict2info(info)
            # format_str += dm.misc.dict2info(info)
            return format_str
        elif ret_fmt == 'dict':
            return info
        elif ret_fmt == 'json':
            return json.dumps(info, indent=4, ensure_ascii=False)
        else:
            raise ValueError(f"Unknown format: {ret_fmt}")

    def dict2info(self, info_dict):
        # 先递归地展开
        new_info_dict = collections.OrderedDict()
        indents = []
        indent_space = 4
        for k, v in info_dict.items():
            self.recursive_func(k, v, new_info_dict, indents, indent_space=indent_space)
        info_dict = new_info_dict

        indents2color = {0: '32m', 1 * indent_space: '35m', 2 * indent_space: '36m', 3 * indent_space: '33m'}
        lenth = np.max([len(x) for x in info_dict.keys()])
        format_str = ''
        for i, (k, v) in enumerate(info_dict.items()):
            indent = indents[i]
            # print(f'k: {k} {len(k)}')
            k_str = f'{k.strip("*"):>{indent + lenth}}'
            color = indents2color.get(indent, None)
            k_str = f'\033[1;{color}{k_str}\033[0m' if color else k_str  # 上色
            format_str += f'  {k_str} : {v}\n'
        return format_str

    def recursive_func(self, k, v, new_info_dict, indents, indent=0, indent_space=4):
        """递归拆分，原来的字典里，某些值可能还是dict，为了方便显示，递归展开，子dict的每个键作为新的顶层的键"""
        if isinstance(v, dict):
            while True:
                if k not in new_info_dict.keys():
                    break
                k = '*' + k
            new_info_dict[f'{k}'] = f'({len(v.keys())})'
            indents.append(indent)
            indent += indent_space
            for k2, v2 in v.items():
                # print(f'k2: {k2} indent: {indent}', new_info_dict.keys())
                self.recursive_func(k2, v2, new_info_dict, indents, indent, indent_space=indent_space)
        else:
            while True:
                if k not in new_info_dict.keys():
                    break
                k = '*' + k
            new_info_dict[f'{k}'] = v
            indents.append(indent)
        return new_info_dict

    def include(self, format=list):
        if format == list:
            return [x['type'] for x in self.modules_list_config]
        else:
            return ' '.join([x['type'] for x in self.modules_list_config])

    def pop(self, wait=True, timeout=None):
        return self.mo.pop(wait=wait, timeout=timeout)

    def get_last(self, wait=True, timeout=None):
        return self.mo.get_last(wait=wait, timeout=timeout)

    def scan(self, last=False):
        return self.mo.scan(last=last)

    def run(self, **kwargs):
        """从stopped运行流"""
        return self.parent.run(self, **kwargs)  # run in main thread

    def start(self, **kwargs):
        return self.parent.run_stream(self, run_in_main_thread=False, **kwargs)

    def evaluate(self, **kwargs):
        return self.parent.evaluate(self, run_in_main_thread=True, **kwargs)

    def train(self, **kwargs):
        return self.parent.train(self, run_in_main_thread=True, **kwargs)

    def infer(self, **kwargs):
        return self.parent.infer(self, run_in_main_thread=True, **kwargs)

    def stop(self):
        if self.status == 'stopped':
            return
        del self._inited_modules  # 引用计数置0，释放内存
        self._inited_modules = collections.OrderedDict()
        self.status = 'stopped'  # 设置成stopped后, self.modules的值就是未出初始化过的模型了



import eventlet
import socketio

debug = False
sio = socketio.Server(logger=debug, engineio_logger=debug)
app = socketio.WSGIApp(sio, static_files={
    '/': {'content_type': 'text/html', 'filename': 'index.html'}
})


@sio.event
def connect(sid, environ):
    print('connect ', sid)
    print('environ', environ)


@sio.event
def my_message(sid, data):
    print(f'my_message received message: {data}')


@sio.event
def server_run_func(sid, data):
    print(f'server run_func, data: {data}')
    return 'dame', 'dame2'


@sio.event
def disconnect(sid):
    print('disconnect ', sid)


if __name__ == '__main__':
    eventlet.wsgi.server(eventlet.listen(('', 5000)), app)




from .. import uaii


def list():
    """
    list all algorithms
    :return: list of algorithms
    """
    info = uaii.ps()
    # print(info)
    return info


def load(name, *args, **kwargs):
    """
    load algorithm by name  
    :param name: algorithm name 
    :return: model
    """
    model = uaii.get_module(name, *args, **kwargs)
    model = model()
    return model
    
def docs(name):
    """
    get algorithm docs by name
    :param name: algorithm name
    :return: docs
    """
    info = f'https://ai.ihep.ac.cn/docs/{name}'
    return info



import os, sys


def get_registed_module_info(registry, module=None, **kwargs):
    """
    ID', 'TYPE', 'NAME', 'STATUS', 'TAG', 'INCLUDE', 'DESCRIPTION'
    分析已注册注册模块的信息，模块的状态存在自己里。
    :param modules: 对象，Registry里的modules_dict
    :param start_id: int，从第start_id个统计模块数目。
    :return: list, (n, 4) n个模块，4: 模块ID, 模块名(内部名)，模块状态，模块描述
    """
    include_dict = kwargs.get('include_dict', None)

    if module is None:
        items = registry.module_dict.items()
        ids = registry.module_ids
    else:
        modules = [module] if isinstance(module, str) else module
        items = dict()
        ids = list()
        for i, (name, module) in enumerate(registry.module_dict.items()):
            if name in modules:
                items[name] = module
                ids.append(registry.module_ids[i])
    meta = []
    for i, (name, module) in enumerate(items):
        # name是注册模块的name
        id = ids[i]
        type = registry.name
        mname = module.name  # 模块内部的name
        mname = name if mname.lower() == 'default_name' else mname  # 如果是默认名字，用注册名替代
        mstatus = module.status  # 状态
        mdescription = module.description  # 描述
        # print(type)
        mtag = module.tag if type == 'module' else '-'  # 标签
        # mtag = module.tag
        # print(f'name: {name}')
        # print(f'module: {module.name}')
        meta.append([id, type, mname, mstatus, mtag, '-', mdescription])

    return meta



"""
初始化注册，
在导入nn模块自动注册内部模块时，由于外部调用多次会导致对此初始化注册，仅初始化注册一次
"""
import os
import sys
from pathlib import Path


class InitRegister(object):
    def __init__(self, internal_dir=None):
        self.inited = False
        self.internal_dir = internal_dir  # /home/zzd/VSProject/hai/hai

    def import_internal_module(self, internal_module_path):
        """
        导入内部模块
        :param internal_module_path: internal module path, e.g.: apis/modules/YOLOv5-v6
        """
        if self.internal_dir is None:
            full_path = os.path.join(os.getcwd(), internal_module_path)
        else:
            full_path = os.path.join(self.internal_dir, internal_module_path)  # /home/zzd/VSProject/hai/hai/apis/modules/YOLOv5_v6
        full_path = Path(full_path)
        # dir_path = str(full_path.parent)
        # stem = str(full_path.stem)

        xx = internal_module_path.split('/')
        xx = '.'.join(xx)

        code = f'from {xx} import __init__'
        exec(code)


    def __call__(self, internal_modules, external_folders):
        """
        初始化注册
        :param internal_modules: internal modules, e.g: models/loader/vis_loader
        :param external_folders: external folders, e.g.: dmapi, repos
        """
        if self.inited:
            pass
        else:
            # print(self.inited)
            ims = internal_modules
            efs = external_folders  # ['dmapi', 'repos']
            if len(ims) > 0:
                for i, im_path in enumerate(ims):
                    self.import_internal_module(im_path)

            if len(efs) > 0:
                dirs = [x for x in os.listdir('.') if os.path.isdir(x)]  # 当前路径文件夹
                for dmp in dirs:
                    if dmp in efs:
                        # code = f'from repos.xxx.dmapi import *'
                        code = f'from {dmp} import *'
                        exec(code)
            self.inited = True




import inspect
import warnings
from functools import partial
import damei as dm
from collections import OrderedDict

from .utils.general import get_registed_module_info


class Registry:
    """A registry to map strings to classes.

    Registered object could be built from registry.

    Example:
        >>> MODELS = Registry('models')
        >>> @MODELS.register_module()
        >>> class ResNet:
        >>>     pass
        >>> resnet = MODELS.build(dict(type='ResNet'))

    Please refer to
    https://mmcv.readthedocs.io/en/latest/understand_mmcv/registry.html for
    advanced usage.

    Args:
        name (str): Registry name.
        build_func(func, optional): Build function to construct instance from
            Registry, func:`build_from_cfg` is used if neither ``parent`` or
            ``build_func`` is specified. If ``parent`` is specified and
            ``build_func`` is not given,  ``build_func`` will be inherited
            from ``parent``. Default: None.
        parent (Registry, optional): Parent registry. The class registered in
            children registry could be built from parent. Default: None.
        scope (str, optional): The scope of registry. It is the key to search
            for children registry. If not specified, scope will be the name of
            the package where class is defined, e.g. mmdet, mmcls, mmseg.
            Default: None.
    """

    def __init__(self, name, build_func=None, parent=None, scope=None, id_prefix=None):
        self._name = name
        self._id_prefix = id_prefix if id_prefix else self._name[0]
        self._module_dict = OrderedDict()
        self._module_ids = []
        self._children = dict()
        self._scope = self.infer_scope() if scope is None else scope

        # self.build_func will be set with the following priority:
        # 1. build_func
        # 2. parent.build_func
        # 3. build_from_cfg
        if build_func is None:
            if parent is not None:
                self.build_func = parent.build_func
            else:
                #
                self.build_func = None
                # self.build_func = build_from_cfg
        else:
            self.build_func = build_func
        if parent is not None:
            assert isinstance(parent, Registry)
            parent._add_children(self)
            self.parent = parent
        else:
            self.parent = None

    def __len__(self):
        return len(self._module_dict)

    def __contains__(self, key):
        return self.get(key) is not None

    def __repr__(self):
        im = dm.misc.list2table([['       ', f'{k}:', str(v)] for k, v in self._module_dict.items()])
        format_str = self.__class__.__name__ + \
                     f'(name={self._name}, ' \
                     f'items=\n{im})'
        return format_str

    @staticmethod
    def infer_scope():
        """Infer the scope of registry.

        The name of the package where registry is defined will be returned.

        Example:
            >>> # in mmdet/models/backbone/resnet.py
            >>> MODELS = Registry('models')
            >>> @MODELS.register_module()
            >>> class ResNet:
            >>>     pass
            The scope of ``ResNet`` will be ``mmdet``.

        Returns:
            str: The inferred scope name.
        """
        # inspect.stack() trace where this function is called, the index-2
        # indicates the frame where `infer_scope()` is called
        filename = inspect.getmodule(inspect.stack()[2][0]).__name__
        # print(f'filename: {filename}')
        split_filename = filename.split('.')
        return split_filename[0]

    @staticmethod
    def split_scope_key(key):
        """Split scope and key.

        The first scope will be split from key.

        Examples:
            >>> Registry.split_scope_key('mmdet.ResNet')
            'mmdet', 'ResNet'
            >>> Registry.split_scope_key('ResNet')
            None, 'ResNet'

        Return:
            tuple[str | None, str]: The former element is the first scope of
            the key, which can be ``None``. The latter is the remaining key.
        """
        split_index = key.find('.')
        if split_index != -1:
            return key[:split_index], key[split_index + 1:]
        else:
            return None, key

    @property
    def name(self):
        return self._name

    @property
    def scope(self):
        return self._scope

    @property
    def module_dict(self):
        return self._module_dict

    @property
    def module_ids(self):
        return self._module_ids

    @property
    def children(self):
        return self._children

    def get(self, key):
        """Get the registry record.

        Args:
            key (str): The class name in string format.

        Returns:
            class: The corresponding class.
        """
        scope, real_key = self.split_scope_key(key)
        if scope is None or scope == self._scope:
            # get from self
            if real_key in self._module_dict:
                return self._module_dict[real_key]
        else:
            # get from self._children
            if scope in self._children:
                return self._children[scope].get(real_key)
            else:
                # goto root
                parent = self.parent
                while parent.parent is not None:
                    parent = parent.parent
                return parent.get(key)

    def build(self, *args, **kwargs):
        return self.build_func(*args, **kwargs, registry=self)

    def _add_children(self, registry):
        """Add children for a registry.

        The ``registry`` will be added as children based on its scope.
        The parent registry could build objects from children registry.

        Example:
            >>> models = Registry('models')
            >>> mmdet_models = Registry('models', parent=models)
            >>> @mmdet_models.register_module()
            >>> class ResNet:
            >>>     pass
            >>> resnet = models.build(dict(type='mmdet.ResNet'))
        """

        assert isinstance(registry, Registry)
        assert registry.scope is not None
        assert registry.scope not in self.children, \
            f'scope {registry.scope} exists in {self.name} registry'
        self.children[registry.scope] = registry

    def _register_module(self, module_class, module_name=None, force=False):
        # print('xx register')
        # exit()
        if not inspect.isclass(module_class):
            raise TypeError('module must be a class, '
                            f'but got {type(module_class)}')

        if module_name is None:
            module_name = module_class.__name__
        if isinstance(module_name, str):
            module_name = [module_name]

        # print(module_name, self._module_dict)
        for name in module_name:
            if not force and name in self._module_dict:
                raise KeyError(f'{name} is already registered '
                               f'in {self.name}')
            self._module_dict[name] = module_class
            if self._module_ids:
                max_id = int(self._module_ids[-1][len(self._id_prefix)::])
                self._module_ids.append(f'{self._id_prefix}{str(max_id + 1):0>2}')
            else:
                self._module_ids.append(f'{self._id_prefix}01')

    def deprecated_register_module(self, cls=None, force=False):
        warnings.warn(
            'The old API of register_module(module, force=False) '
            'is deprecated and will be removed, please use the new API '
            'register_module(name=None, force=False, module=None) instead.',
            DeprecationWarning)
        if cls is None:
            return partial(self.deprecated_register_module, force=force)
        self._register_module(cls, force=force)
        return cls

    def register_module(self, name=None, force=False, module=None):
        """Register a module.

        A record will be added to `self._module_dict`, whose key is the class
        name or the specified name, and value is the class itself.
        It can be used as a decorator or a normal function.

        Example:
            >>> backbones = Registry('backbone')
            >>> @backbones.register_module()
            >>> class ResNet:
            >>>     pass

            >>> backbones = Registry('backbone')
            >>> @backbones.register_module(name='mnet')
            >>> class MobileNet:
            >>>     pass

            >>> backbones = Registry('backbone')
            >>> class ResNet:
            >>>     pass
            >>> backbones.register_module(ResNet)

        Args:
            name (str | None): The module name to be registered. If not
                specified, the class name will be used.
            force (bool, optional): Whether to override an existing class with
                the same name. Default: False.
            module (type): Module class to be registered.
        """
        # print(f'reg: {name}')
        if not isinstance(force, bool):
            raise TypeError(f'force must be a boolean, but got {type(force)}')
        # NOTE: This is a walkaround to be compatible with the old api,
        # while it may introduce unexpected bugs.
        if isinstance(name, type):
            return self.deprecated_register_module(name, force=force)

        # raise the error ahead of time
        if not (name is None or isinstance(name, str) or is_seq_of(name, str)):
            raise TypeError(
                'name must be either of None, an instance of str or a sequence'
                f'  of str, but got {type(name)}')

        # use it as a normal method: x.register_module(module=SomeClass)
        if module is not None:
            self._register_module(
                module_class=module, module_name=name, force=force)
            return module

        # use it as a decorator: @x.register_module()
        def _register(cls):
            self._register_module(
                module_class=cls, module_name=name, force=force)
            return cls

        return _register

    @property
    def name2id_dict(self):
        """返回所有的本注册器里模块名为键，id为值组成的list"""
        _name2id_dict = dict()
        for i, (name, module) in enumerate(self.module_dict.items()):
            _name2id_dict[name] = self.module_ids[i]
        return _name2id_dict

    def ps(self, module=None, **kwargs):
        """获取本注册器里各个模块的信息"""
        # print(self.module_dict)
        # print(self.module_ids)
        # exit()
        return get_registed_module_info(self, module=module, **kwargs)



"""
damei nn api for ai algorithm management (unified ai interface)
"""

from ..version import __backend__, __version__
import os, sys
from pathlib import Path

# damei.nn warpper
__pydir__ = Path(os.path.dirname(os.path.abspath(__file__)))

# 导入damei lib
try:
    import damei
except:
    sys.path.append(f'{__pydir__.parent}/damei')
    import damei

logger = damei.getLogger('hai')
logger.info(f'HAI version: {__version__}')


if __backend__ == 'local':
    from .basic.base import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
    from .basic.registry import MODULES, SCRIPTS, IOS, init_register
    from .basic.utils import Config
    from ..uaii.uaii_main import UAII
    from ..tests.test import test_module, test_script, test_io

elif __backend__ == 'damei':
    from damei.nn.api.base import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
    from damei.nn.api.registry import MODULES, SCRIPTS, IOS
    from damei.nn.api.utils import Config
    from damei.nn.api.utils import Config as PyConfigLoader
    from damei.nn.api import UAII as UAII
else:
    raise NotImplementedError(f'{__backend__} backend is not supported, only local and damei, please check')

root_path = f'{Path(__pydir__).parent.parent}'
uaii = UAII()


"""
为了任意路径下的.py包的引入
"""
import os, sys
from pathlib import Path
import damei as dm

logger = dm.getLogger('arbitray_import')


def arbitrary_import(path):
    logger.warn(f'Not tested.')
    if os.path.isdir(path):
        path = f'{path}/__init__.py'
    abs_path = Path(os.path.abspath(path))
    cdir = os.getcwd()  # 当前工作目录
    sys.path.append(f'{abs_path.parent}')
    module_dir = '.'
    os.chdir(f'{abs_path.parent}')  # 切换工作目录
    code = f'import {abs_path.stem}'

    try:
        exec(code)
    except Exception as e:
        msg = f'Error in "exec(code)": {e}. Error variables:\n' \
              f'    abs path: {abs_path} \n' \
              f'    code: {code} \n' \
              f'    module dir: {module_dir} \n' \
              f'    sys.path[-1]: {sys.path[-1]} '
        raise Exception(msg)

    sys.path = sys.path[:-1:]
    os.chdir(cdir)





from .apis import AbstractInput, AbstractModule, AbstractOutput, AbstractQue
from .apis import MODULES, SCRIPTS, IOS, init_register
from .apis import Config
from .apis import UAII, uaii
from .apis import hub
from .apis import root_path

# from xsensing_ai.modules import *  # 加载项目的模块
# from xsensing_ai.uaii.server.grpc.grpc_xai_client import XAIGrpcClient



# """
# 导入内部模块和外部模块
internal_modules = [
    'hai/apis/modules/YOLOv5',
    'hai/apis/modules/ResNet',
    'hai/apis/modules/UNet',
    'hai/modules/loader/images_loader',
    'hai/modules/exporter/images_exporter',
]
external_folders = ['dmapi', 'repos']

init_register(internal_modules=internal_modules, external_folders=external_folders)
# """






"""
提供基本的UAII类，把各种函数封装在这里。
"""
import collections
import os, sys
import numpy as np
import damei as dm
import cv2
import json

logger = dm.getLogger('base_uaii')
from . import general


class BaseUAII(object):
    def __init__(self, **kwargs):
        self._modules = None
        self._scripts = None
        self._ios = None
        self._streams = None
        self._name2cls = None
        self.cfg = None
        self.is_center = kwargs.get('is_center', True)

    def _configure(self):
        """区分云边端模型"""

        is_center = self.is_center
        if is_center:
            return
        # 如果是边缘端，只保留tag为edge的模型
        ms = self.modules  # Registry对象

        valid_ikv = [[i, k, v] for i, (k, v) in enumerate(ms._module_dict.items()) if v.tag == 'edge']

        new_module_dict = collections.OrderedDict()
        new_module_ids = list()
        for i, k, v in valid_ikv:
            new_module_dict[k] = v
            new_module_ids.append(ms._module_ids[i])

        ms._module_dict = new_module_dict
        ms._module_ids = new_module_ids
        # print(ms._module_dict, len(ms._module_dict))
        # print(ms._module_ids)
        # exit()
        # pass

    def register_name2cls(self):
        name2cls = dict()
        for k, v in self.modules.module_dict.items():
            name = k
            name2cls[name] = 'MODULE'
        for k, v in self.scripts.module_dict.items():
            name2cls[k] = 'SCRIPT'
        for k, v in self.streams.module_dict.items():
            name2cls[k] = 'STREAM'
        for k, v in self.ios.module_dict.items():
            name2cls[k] = 'IO'
        return name2cls

    @property
    def name2cls(self):
        if self._name2cls is None or self._name2cls == dict():
            # print('Register name2cls...')
            self._name2cls = self.register_name2cls()
        return self._name2cls

    @property
    def modules(self):
        return self._modules

    @property
    def scripts(self):
        return self._scripts

    @property
    def streams(self):
        return self._streams

    @property
    def ios(self):
        return self._ios

    def get_module(self, name, cls=None, *args, **kwargs):
        """根据模块名和类别获取模块，是未初始化的"""
        # print(args, kwargs, cls)
        ps = self.ps(**kwargs)
        # 输入的名字也可以是ID
        exist_names = [x.split()[2] for x in ps.split('\n')[1::]]
        if name not in exist_names:
            exist_ids = [x.split()[0] for x in ps.split('\n')[1::]]
            assert len(exist_ids) == len(exist_names)
            if name in exist_ids:
                name = exist_names[exist_ids.index(name)]
            else:
                raise ModuleNotFoundError(f"Expected module name: '{name}' not found in exist modules: {exist_names}.")

        # try:
        cls = cls if cls else self.name2cls[name]
        # except Exception as e:
            # raise ModuleNotFoundError(
                # f'Error: {e}. \nAll modules: {self.modules}.')

        # print(cls)
        if cls == 'MODULE':
            return self.modules.get(name)
        elif cls == 'SCRIPT':
            return self.scripts.get(name)
        elif cls == 'STREAM':
            return self.streams.get(name)
        elif cls == 'IO':
            return self.ios.get(name)
        else:
            raise NotImplementedError(f'UAII get module name: {name} cls: {cls} not implemented.')

    def build_stream(self, cfg, **kwargs):
        """
        根据流配置文件创建新的流，注册到streams里
        cfg: stream config, or module_name
            for example:
            >>> cfg = dict(
            >>>    type='radar_detection_stream',
            >>>    description='雷达目标探测流',
            >>>    models=[
            >>>        dict(type='radardet', cfg=None)
            >>>    ])
            for example2:
            >>> cfg = 'radardet'
        return: stream object
        """
        is_req = kwargs.get('is_req', False)  # 是否gRPC的请求，如果是，返回值需要重新定义，不能返回对象
        name = f'{cfg}_stream' if isinstance(cfg, str) else cfg['type']
        # 判断算法流是否已经存在
        if name in self.streams.names:
            logger.warn(f'Build stream, name:{name} exist, return existing stream.')
            if is_req:
                return 0, f'stream: {name} exist.'  # 失败0，成功1
            else:
                return self.get_stream(name=name)

        # 创建新的算法流
        if isinstance(cfg, str):  # 就自动配置一个单模块流
            m_name = cfg
            module_desc = self.get_module(m_name).description
            cfg = dict(
                type=f'{m_name}_stream',
                description=f'mono {module_desc} stream',
                models=[
                    dict(
                        type=m_name,
                    )
                ]
            )
        # print(cfg)
        self.streams.build_new_stream(cfg, **kwargs)

        if is_req:
            return 1, name
        else:
            return self.get_stream(name)

    def get_stream(self, name):
        """获得存在的流，并初始化内部的所有模块"""
        stream = self.get_module(name=name, cls='STREAM')
        return stream

    def get_stream_info(self, stream=None, stream_name=None, **kwargs):
        """获得流的信息"""
        is_req = kwargs.pop('is_req', False)
        stream = stream if stream else stream_name
        stream = self.get_stream(stream) if isinstance(stream, str) else stream
        assert stream is not None, f'Stream: {stream} not found.'
        if is_req:
            return 1, stream.info(**kwargs)
        return stream.info(**kwargs)

    def get_stream_cfg_meta(self, stream=None, stream_name=None, addr='/', **kwargs):
        """获得流的配置文件的元信息"""
        is_req = kwargs.pop('is_req', False)
        ret_fmt = kwargs.pop('ret_fmt', 'Config object')

        stream = stream if stream else stream_name
        stream = self.get_stream(stream) if isinstance(stream, str) else stream
        assert stream is not None, f'Stream: {stream} not found.'
        config_meta = stream.get_cfg_meta(addr=addr, **kwargs)
        if is_req:
            return 1, config_meta.to_json()
        else:
            if ret_fmt == 'Config object':
                return config_meta
            elif ret_fmt == 'dict':
                return config_meta.to_dict()
            elif ret_fmt == 'json':
                return config_meta.to_json()
            else:
                raise NotImplementedError(f'UAII get_stream_cfg_meta ret_fmt: {ret_fmt} not implemented.')

    def get_stream_cfg(self, stream=None, stream_name=None, addr='/', **kwargs):
        """
        获得流的配置参数
        """
        is_req = kwargs.pop('is_req', False)  # 是否gRPC的请求，如果是，返回值需要重新定义，
        ret_fmt = kwargs.pop('ret_fmt', 'Config object')  # 返回格式，默认是xai.Config对象
        assert ret_fmt in ['Config object', 'dict'], f'Ret format: {ret_fmt} not supported.'

        stream = stream if stream else stream_name
        stream = self.get_stream(stream) if isinstance(stream, str) else stream
        assert stream is not None, f'Stream: {stream} not found.'
        config = stream.get_cfg(addr=addr, **kwargs)
        # print(config)
        # logger.info(f'Get stream config: {config} {type(config)}')
        if is_req:
            return 1, config.to_json()
        else:
            if ret_fmt == 'Config object':
                return config
            return config.to_json()

    def get_stream_status(self, stream, **kwargs):
        """获得流的状态"""
        is_req = kwargs.pop('is_req', False)
        stream = self.get_stream(stream) if isinstance(stream, str) else stream
        assert stream is not None, f'Stream: {stream} not found.'
        if is_req:
            return 1, stream.status
        return stream.status

    def set_stream_cfg(self, cfg, stream=None, stream_name=None, addr='/', **kwargs):
        """
        设置流的参数
        :param cfg:  流的参数配置，dict, list, str, int, float, bool, etc.
        :param stream: 流对象，如果为None，则根据stream_name获取流对象, xai.Stream对象
        :param stream_name: 流名称，str
        :param addr: 需要修改的流配置的的地址，str
            for example:
                addr = '/'  # 表示修改整个配置
                addr = '/model_name'  # 表示修改流内某个model配置
                addr = '/model_name/param_name'  # 表示修改流内某个model的具体某个param配置，以此类推
        :return: uaii调用时返回xai.Config对象
                grpc调用时返回(1, 字典类型的配置) or (0, 错误信息)
        """
        is_req = kwargs.pop('is_req', False)  # 是否gRPC的请求，如果是，返回值需要重新定义，不能返回对象
        stream = stream if stream else stream_name
        stream = self.get_stream(stream) if isinstance(stream, str) else stream
        assert stream is not None, f'Stream not found. {stream}'
        config = stream.set_cfg(addr=addr, value=cfg, **kwargs)
        # print(config)
        if is_req:
            return 1, config.to_json()
        else:
            return config

    def get_model(self, name):
        """读取单独的模块，内部是把单独的模块组装成流"""
        exists_streams = self.streams.names
        if name in exists_streams:
            mono_flow = self.get_module(name=name, cls='STREAM')
        else:
            cfg = dict(
                type=f'{name}_flow',
                description=f'mono module flow: {name}',
                models=[
                    dict(type=name, cfg=None),
                ]
            )
            mono_flow = self.build_stream(cfg=cfg)
        return mono_flow

    def stream_info(self, name):
        return self.get_stream(name=name).info()

    def ps(self, stream=None, module=None, io=None, script=None, **kwargs):
        """返回全部模块的状态"""
        ret_fmt = kwargs.get('ret_fmt', 'string')
        xai_type = kwargs.get('type', None)  # xai_type: 'stream', 'module', 'io', 'script'
        if xai_type is None:
            pass
        else:
            if xai_type == 'stream':
                stream = True
            elif xai_type == 'module':
                module = True
            elif xai_type == 'io':
                io = True
            elif xai_type == 'script':
                script = True
            else:
                raise ValueError(f'xai_type: {xai_type} not supported.')

        ps0 = [['ID', 'TYPE', 'NAME', 'STATUS', 'TAG', 'INCLUDE', 'DESCRIPTION']]
        if stream is None and module is None and io is None and script is None:
            tmp = self.streams.ps(include_dict=self.modules.name2id_dict)
            ps0 += tmp
            tmp2 = self.modules.ps(include_dict=self.ios.name2id_dict)
            ps0 += tmp2
            ps0 += self.ios.ps()
            # print(ps0, len(ps0))
            # ps0.append(self.scripts.ps())
        else:
            if stream:
                tmp = self.streams.ps(include_dict=self.modules.name2id_dict)
                ps0 += tmp
            if module:
                tmp = self.modules.ps(include_dict=self.modules.name2id_dict)
                ps0 += tmp
            if io:
                ps0 += self.ios.ps(module=None)
            if script:
                ps0 += self.scripts.ps(module=None)
        # ps_all = ps0
        # lenth = np.array([[len(x) for x in xx] for xx in ps_all])  # (n, 4)
        # print()
        # lenth = np.max(lenth, axis=0)  # (4,)
        # ps_list = [[f'{x:<{lenth[i]}}' for i, x in enumerate(xx)] for xx in ps_all]
        # ps_str = '\n'.join(['  '.join(x) for x in ps_list])
        if ret_fmt == 'string':
            return dm.misc.list2table(ps0)
        elif ret_fmt == 'json':
            return json.dumps(ps0, indent=4, ensure_ascii=False)
        elif ret_fmt == 'list':
            return ps0
        else:
            raise NotImplementedError(f'{ret_fmt} not implemented.')

    def show_vis(self, ret, im0, target_names):

        return general.single_plot(ret, im0, target_names=target_names)
    



from .registy import Registry
from .init_register import InitRegister
import os
from pathlib import Path
pydir = Path(os.path.abspath(__file__)).parent

# print('registry init')

MODULES = Registry('module', build_func=None)
SCRIPTS = Registry('script', build_func=None)
# STREAMS = Registry('stream', build_func=None)
IOS = Registry('io', build_func=None, id_prefix='IO')

init_register = InitRegister(internal_dir=f'{pydir.parent.parent.parent}')
